(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 6.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[     68455,       1877]
NotebookOptionsPosition[     26864,        918]
NotebookOutlinePosition[     64131,       1731]
CellTagsIndexPosition[     64088,       1728]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{
Cell["Large Expression", "Title"],

Cell["Code Generation for a large expression", "Subtitle"],

Cell["\<\
Copyright \[Copyright] MathCore AB 1998-2004\
\>", "Copyright"],

Cell[CellGroupData[{

Cell["Subexpression Elimination", "Section"],

Cell["\<\
Common subexpression elimination on symbolic expressions is used for the \
following reasons:\
\>", "Text"],

Cell["Generating code from LARGE expressions", "Itemize"],

Cell["Performance Optimization", "Itemize"],

Cell[CellGroupData[{

Cell["A Large (Realistic) Example", "Subsection"],

Cell[TextData[{
 " Compute a ",
 StyleBox["rigidity matrix",
  FontWeight->"Bold"],
 " and second member  for a finite element model of a geometrically nonlinear \
",
 StyleBox["3D beam",
  FontWeight->"Bold"],
 " element. (Example from Pierre Aubert ",
 StyleBox["paubert@mathinsa.insa-lyon.fr",
  FontSlant->"Italic"],
 ")"
}], "Text",
 CellChangeTimes->{3.508661950561092*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"Off", "[", 
   RowBox[{"General", "::", "spell1"}], "]"}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell["The Energy Expression", "Subsubsection"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"If", " ", "[", 
    RowBox[{
     RowBox[{"$VersionNumber", " ", "<", "6"}], ",", "   ", 
     RowBox[{
     "Needs", "[", "\"\<LinearAlgebra`MatrixManipulation`\>\"", "]"}]}], 
    "]"}], ";"}], "\n", "\n", 
  RowBox[{"(*", "                                       ", "*)"}], "\n", 
  RowBox[{"(*", " ", 
   RowBox[{"shape", " ", "function"}], "                       ", "*)"}], 
  "\n", "\n", 
  RowBox[{"(*", "                                       ", "*)"}]}], "\n", 
 RowBox[{
  RowBox[{"Ni", " ", "=", " ", 
   RowBox[{"1", "-", 
    RowBox[{"x", "/", "l"}]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"Nj", " ", "=", 
    RowBox[{"x", "/", "l"}]}], ";"}], "\n", "\n", 
  RowBox[{"(*", "                                      ", "*)"}], "\n", 
  RowBox[{"(*", " ", 
   RowBox[{"interpolations", " ", "and", " ", "derivatives"}], "         ", 
   "*)"}], "\n", 
  RowBox[{"(*", "                                        ", "*)"}]}], "\n", 
 RowBox[{
  RowBox[{"u1x", " ", "=", " ", 
   RowBox[{
    RowBox[{"Ni", "*", "u1i"}], " ", "+", " ", 
    RowBox[{"Nj", "*", "u1j"}]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"u2x", " ", "=", " ", 
   RowBox[{
    RowBox[{"Ni", "*", "u2i"}], " ", "+", " ", 
    RowBox[{"Nj", "*", "u2j"}]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"u3x", " ", "=", " ", 
   RowBox[{
    RowBox[{"Ni", "*", "u3i"}], " ", "+", " ", 
    RowBox[{"Nj", "*", "u3j"}]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"t1x", " ", "=", " ", 
   RowBox[{
    RowBox[{"Ni", "*", "t1i"}], " ", "+", " ", 
    RowBox[{"Nj", "*", "t1j"}]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"t2x", " ", "=", " ", 
   RowBox[{
    RowBox[{"Ni", "*", "t2i"}], " ", "+", " ", 
    RowBox[{"Nj", "*", "t2j"}]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"t3x", " ", "=", " ", 
    RowBox[{
     RowBox[{"Ni", "*", "t3i"}], " ", "+", " ", 
     RowBox[{"Nj", "*", "t3j"}]}]}], ";"}], "\n"}], "\n", 
 RowBox[{
  RowBox[{"du1", "=", 
   RowBox[{"D", "[", 
    RowBox[{"u1x", ",", "x"}], "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"du2", "=", 
   RowBox[{"D", "[", 
    RowBox[{"u2x", ",", "x"}], "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"du3", "=", 
   RowBox[{"D", "[", 
    RowBox[{"u3x", ",", "x"}], "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"dt1", "=", 
   RowBox[{"D", "[", 
    RowBox[{"t1x", ",", "x"}], "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"dt2", "=", 
   RowBox[{"D", "[", 
    RowBox[{"t2x", ",", "x"}], "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"dt3", "=", 
    RowBox[{"D", "[", 
     RowBox[{"t3x", ",", "x"}], "]"}]}], ";"}], "\n", "\n", 
  RowBox[{"(*", "                                     ", "*)"}], "\n", 
  RowBox[{"(*", " ", 
   RowBox[{
   "compute", " ", "a", " ", "closed", " ", "form", " ", "of", " ", "the", 
    " ", "expential", " ", "matrix"}], " ", "*)"}], "\n", 
  RowBox[{
  "(*", "                                            ", "*)"}]}], "\n", 
 RowBox[{
  RowBox[{"al", "=", 
   RowBox[{"Sqrt", "[", 
    RowBox[{
     RowBox[{"t1x", "^", "2"}], "+", 
     RowBox[{"t2x", "^", "2"}], "+", 
     RowBox[{"t3x", "^", "2"}]}], "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"a", "  ", "=", 
   RowBox[{
    RowBox[{"Sin", "[", "al", "]"}], "/", "al"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"a1", "=", 
   RowBox[{"1", "-", "a"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"a2", "=", 
   RowBox[{
    RowBox[{"1", "/", "2"}], "*", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"Sin", "[", 
        RowBox[{"al", "/", "2"}], "]"}], "/", 
       RowBox[{"(", 
        RowBox[{"al", "/", "2"}], ")"}]}], ")"}], "^", "2"}]}]}], "\t", 
  ";"}], "\n", 
 RowBox[{
  RowBox[{"x1", "=", 
   RowBox[{
    RowBox[{"1", "/", "2"}], "*", "l"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"theta", "=", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", 
       RowBox[{"-", "t3x"}], ",", "t2x"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"t3x", ",", "0", ",", 
       RowBox[{"-", "t1x"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "t2x"}], ",", "t1x", ",", "0"}], "}"}]}], "}"}]}], 
  ";"}], "\n", 
 RowBox[{
  RowBox[{"vec", "    ", "=", " ", 
   RowBox[{"{", 
    RowBox[{"t1x", ",", "t2x", ",", "t3x"}], "}"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"uni", " ", "=", " ", 
   RowBox[{"{", 
    RowBox[{"1", ",", "0", ",", "0"}], "}"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"du", "=", 
   RowBox[{"{", 
    RowBox[{"du1", ",", "du2", ",", "du3"}], "}"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"dt", "=", 
    RowBox[{"{", 
     RowBox[{"dt1", ",", "dt2", ",", "dt3"}], "}"}]}], ";"}], "\n", "\n", 
  RowBox[{"(*", "                               ", "*)"}], "\n", 
  RowBox[{"(*", " ", "suite", "                         ", "*)"}], "\n", 
  RowBox[{"(*", "                               ", "*)"}]}], "\n", 
 RowBox[{
  RowBox[{"rexp", "=", 
   RowBox[{
    RowBox[{"IdentityMatrix", "[", "3", "]"}], "+", 
    RowBox[{"a", " ", 
     RowBox[{"Transpose", "[", "theta", "]"}]}], "+", 
    RowBox[{"a2", " ", 
     RowBox[{"theta", ".", "theta"}]}]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"tenseurmat", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"rexp", "-", 
       RowBox[{"IdentityMatrix", "[", "3", "]"}]}], ")"}], ".", "uni"}], "+", 
    
    RowBox[{"rexp", ".", "du"}]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"texp", "=", 
   RowBox[{
    RowBox[{"a", " ", 
     RowBox[{"IdentityMatrix", "[", "3", "]"}]}], "+", 
    RowBox[{"Outer", "[", 
     RowBox[{"Times", ",", "vec", ",", "vec"}], "]"}], "+", 
    RowBox[{"a2", " ", 
     RowBox[{"Transpose", "[", "theta", "]"}]}]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"tenseurrot", " ", "=", 
    RowBox[{"texp", ".", "dt"}]}], ";"}], "\n", "\n", 
  RowBox[{"(*", "                                       ", "*)"}], "\n", 
  RowBox[{"(*", " ", 
   RowBox[{"Elasticity", " ", "tensor"}], "                     ", "*)"}], 
  "\n", 
  RowBox[{"(*", "                                       ", "*)"}]}], "\n", 
 RowBox[{
  RowBox[{"ElastMat", "=", 
   RowBox[{"DiagonalMatrix", "[", 
    RowBox[{"{", 
     RowBox[{"EA", ",", "GAy", ",", "GAz"}], "}"}], "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"ElastRot", "=", 
    RowBox[{"DiagonalMatrix", "[", 
     RowBox[{"{", 
      RowBox[{"EIx", ",", "EIy", ",", "EIz"}], "}"}], "]"}]}], ";"}], 
  "\n"}], "\n", 
 RowBox[{
  RowBox[{"Energie", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"1", "/", "2"}], " ", 
     RowBox[{"tenseurmat", ".", "ElastMat", ".", "tenseurmat"}]}], " ", "+", 
    RowBox[{
     RowBox[{"1", "/", "2"}], "  ", 
     RowBox[{"tenseurrot", ".", " ", "ElastRot", ".", "tenseurrot"}]}]}]}], 
  ";"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Energie", " ", "//", " ", "Short"}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{
   RowBox[{
    FractionBox["1", "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"EIx", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"-", 
              FractionBox["t3i", "l"]}], "+", 
             FractionBox["t3j", "l"]}], ")"}], " ", 
           RowBox[{"(", 
            RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], ")"}]}], 
          "+", 
          RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], "+", 
          RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}]}], ")"}], 
        "2"]}], "+", 
      RowBox[{"EIy", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], ")"}], "2"]}], 
      "+", 
      RowBox[{"EIz", " ", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], ")"}], 
        "2"]}]}], ")"}]}], "+", 
   RowBox[{
    FractionBox["1", "2"], " ", 
    RowBox[{"(", 
     RowBox[{"\[LeftSkeleton]", "1", "\[RightSkeleton]"}], ")"}]}]}],
  Short]], "Output",
 CellChangeTimes->{3.5086614580205216`*^9, 3.5086618810619817`*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The second memeber element", "Subsubsection"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", "                                      ", "*)"}], "\n", 
  RowBox[{"(*", "  ", 
   RowBox[{"vector", " ", "of", " ", "ddl"}], "                       ", 
   "*)"}], "\n", 
  RowBox[{"(*", "                                      ", "*)"}], "\n", 
  RowBox[{
   RowBox[{
    RowBox[{"ddl", "=", 
     RowBox[{"{", 
      RowBox[{
      "u1i", ",", "u2i", ",", "u3i", ",", "t1i", ",", "t2i", ",", "t3i", ",", 
       "u1j", ",", "u2j", ",", "u3j", ",", "t1j", ",", "t2j", ",", "t3j"}], 
      "}"}]}], ";"}], "\n"}]}]], "Input",
 CellChangeTimes->{{3.508661625315255*^9, 3.508661628283967*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", "                                      ", "*)"}], "\n", 
  RowBox[{"(*", " ", 
   RowBox[{"second", " ", "member", " ", "F"}], "                      ", 
   "*)"}], "\n", 
  RowBox[{"(*", "                                      ", "*)"}], "\n", 
  RowBox[{
   RowBox[{"F", "=", 
    RowBox[{"Table", "[", 
     RowBox[{
      RowBox[{"D", "[", 
       RowBox[{"Energie", ",", 
        RowBox[{"ddl", "[", 
         RowBox[{"[", "i", "]"}], "]"}]}], "]"}], ",", 
      RowBox[{"{", 
       RowBox[{"i", ",", "1", ",", 
        RowBox[{"Length", "[", "ddl", "]"}]}], "}"}]}], "]"}]}], 
   ";"}]}]], "Input"],

Cell["To print this is not feasible. Compute the size.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ByteCount", "[", "F", "]"}]], "Input"],

Cell[BoxData["3936024"], "Output",
 CellChangeTimes->{3.5086614580986457`*^9, 3.5086618811401052`*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Generate Code", "Subsection"],

Cell["Define Ffunction as the first 4 elements of F", "Text"],

Cell[CellGroupData[{

Cell[TextData[{
 "Needs[\"",
 StyleBox["MathCode`",
  FontColor->RGBColor[0, 0, 1]],
 "\"]"
}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"MathCode C++ 1.4.4\"\>", "\[InvisibleSpace]", "\<\" for \"\>", 
   "\[InvisibleSpace]", "\<\"vc100\"\>", 
   "\[InvisibleSpace]", "\<\" loaded from \"\>", 
   "\[InvisibleSpace]", "\<\"C:\\\\b\\\\omath\\\\MathCode\"\>"}],
  SequenceForm[
  "MathCode C++ 1.4.4", " for ", "vc100", " loaded from ", 
   "C:\\b\\omath\\MathCode"],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086618815463505`*^9}]
}, Open  ]],

Cell["This is needed if many notebooks run in a sequence", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCurrentPackage", "[", "\"\<Global\>\"", "]"}], " "}]], "Input"],

Cell[BoxData["\<\"Global\"\>"], "Output",
 CellChangeTimes->{3.50866145811427*^9, 3.5086618816088495`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Evaluate[
\tFfunction[
\t\tReal EA_,Real EIx_,Real EIy_,Real EIz_,
\t\tReal GAy_,Real GAz_,
\t\tReal l_,
\t\tReal t1i_,Real t1j_,Real t2i_,Real t2j_,Real t3i_,Real t3j_,
\t\tReal u1i_,Real u1j_,Real u2i_,Real u2j_,Real u3i_,Real u3j_,
\t\tReal x_] -> Real[_]] = Take[F,4]\
\>", "Input"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"General", "::", "spell1"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"New symbol name \\\"\[NoBreak]\\!\\(Ffunction\\)\[NoBreak]\
\\\" is similar to existing symbol \
\\\"\[NoBreak]\\!\\(Function\\)\[NoBreak]\\\" and may be misspelled. \
\\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\\\", \
ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/General/spell1\\\", \
ButtonNote -> \\\"General::spell1\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.508661881796347*^9}]
}, Open  ]],

Cell["\<\
To be able to measaure the perfomance we need function looping over the \
Ffunction\
\>", "Text"],

Cell["SetDirectory[$MCRoot<>\"/Demos/LargeExpression\"];", "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"FfunLoop", "[", 
     RowBox[{"Integer", " ", "n_"}], "]"}], "->", 
    RowBox[{"Real", "[", "_", "]"}]}], " ", ":=", " ", "\n", "\t", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"Real", "[", "_", "]"}], "  ", "res"}], "}"}], ",", "\n", 
     "\t\t", 
     RowBox[{
      RowBox[{"Print", "[", "n", "]"}], ";", "\n", "\t\t", 
      RowBox[{"Do", "[", 
       RowBox[{
        RowBox[{"res", " ", "=", 
         RowBox[{"Ffunction", "[", "\[IndentingNewLine]", 
          RowBox[{
          "0.65", ",", "0.65", ",", "0.65", ",", "0.65", ",", 
           "\[IndentingNewLine]", "0.65", ",", "0.65", ",", "0.65", ",", 
           "0.65", ",", "\[IndentingNewLine]", "0.65", ",", "0.65", ",", 
           "0.65", ",", "0.65", ",", "\[IndentingNewLine]", "0.65", ",", 
           "0.65", ",", "0.65", ",", "0.65", ",", "\[IndentingNewLine]", 
           "0.65", ",", "0.65", ",", "0.65", ",", "0.65"}], 
          "\[IndentingNewLine]", "]"}]}], ",", 
        RowBox[{"{", "n", "}"}]}], "]"}], ";", "\n", "\t\t", "res"}]}], "\n", 
    "\t\t", "]"}]}], ";"}]], "Input"],

Cell["\<\
Generate, compile and link. This might take some time, if your computer does \
not have that much of available memory. It is best to free up some.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Share[]", "Input"],

Cell[BoxData["2177136"], "Output",
 CellChangeTimes->{3.5086614583798923`*^9, 3.5086618819994698`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BuildCode", "[", 
  RowBox[{"EvaluateFunctions", "->", 
   RowBox[{"{", "Ffunction", "}"}]}], "]"}]], "Input"],

Cell[BoxData["\<\"Successful compilation to C++: 2 function(s)\"\>"], "Print",
 CellChangeTimes->{3.5086614588017616`*^9, 3.5086619104209805`*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Take Look at the Code", "Subsection"],

Cell["Do this with an editor since the code is large", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compare Performance", "Subsection"],

Cell["Restore some memory", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Share", "[", "]"}]], "Input"],

Cell[BoxData["43240"], "Output",
 CellChangeTimes->{3.5086614907076035`*^9, 3.5086619141240587`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Normal ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " Evaluation"
}], "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Ffunction", "[", 
   RowBox[{"Sequence", "@@", " ", 
    RowBox[{"Table", "[", 
     RowBox[{"0.65", ",", 
      RowBox[{"{", "20", "}"}]}], "]"}]}], "]"}], " ", "//", " ", 
  "Timing"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.`", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.37971682940523654`"}], ",", "0.7109900453920845`", ",", 
     RowBox[{"-", "0.3312732159868481`"}], ",", "0.`"}], "}"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{3.5086614907388525`*^9, 3.5086619142021823`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"mathtime", " ", "=", " ", 
  RowBox[{"%", "[", 
   RowBox[{"[", "1", "]"}], "]"}]}]], "Input"],

Cell[BoxData["0.`"], "Output",
 CellChangeTimes->{3.5086614907544775`*^9, 3.508661914233432*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"FfunLoop", "[", "30", "]"}], "//", "Timing"}]}]], "Input"],

Cell[BoxData["30"], "Print",
 CellChangeTimes->{3.5086614907701025`*^9, 3.508661914249057*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.2809999999999997`", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.37971682940523654`"}], ",", "0.7109900453920845`", ",", 
     RowBox[{"-", "0.3312732159868481`"}], ",", "0.`"}], "}"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{3.5086614910513487`*^9, 3.5086619145146785`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{"mathtime", " ", "=", " ", 
   RowBox[{
    RowBox[{"%", "[", 
     RowBox[{"[", "1", "]"}], "]"}], "/", "30"}]}]}]], "Input"],

Cell[BoxData["0.009366666666666657`"], "Output",
 CellChangeTimes->{3.5086614910669737`*^9, 3.508661914530303*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["External Computation", "Subsubsection"],

Cell["Load the externaly compiled code", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InstallCode", "[", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.508661491301346*^9, 3.508661914670926*^9}]
}, Open  ]],

Cell["Create a new timing function", "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"SetAttributes", "[", 
    RowBox[{"AbsTime", ",", "HoldFirst"}], "]"}], ";"}], "\n"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"AbsTime", "[", "x_", "]"}], " ", ":=", " ", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", "\n", "  ", 
      RowBox[{"start", ",", "res"}], "\n", " ", "}"}], ",", "\n", "   ", 
     RowBox[{
      RowBox[{"start", " ", "=", " ", 
       RowBox[{"AbsoluteTime", "[", "]"}]}], ";", "\n", "   ", 
      RowBox[{"res", "=", "x"}], ";", "\n", "   ", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           RowBox[{"AbsoluteTime", "[", "]"}], "-", "start"}], ")"}], " ", 
         "Second"}], ",", "res"}], "}"}]}]}], "\n", "]"}]}], ";"}]}], "Input"],

Cell[TextData[{
 "To remove the influence the ",
 StyleBox["MathLink",
  FontSlant->"Italic"],
 " duration we compute the "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsTime", "[", 
  RowBox[{"FfunLoop", "[", "2000000", "]"}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"2.8437136`7.905430848492213", " ", "Second"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.37971682940523666`"}], ",", "0.7109900453920848`", ",", 
     RowBox[{"-", "0.33127321598684817`"}], ",", "0.`"}], "}"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{3.5086614941919336`*^9, 3.508661917530265*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"exttime", " ", "=", " ", 
  RowBox[{
   RowBox[{"%", "[", 
    RowBox[{"[", "1", "]"}], "]"}], " ", "/", " ", "2000000"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"1.4218568000000000000184`7.905430848492214*^-6", " ", 
  "Second"}]], "Output",
 CellChangeTimes->{3.5086614942388077`*^9, 3.508661917577139*^9}]
}, Open  ]],

Cell["How much faster?", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"mathtime", " ", "/", " ", "exttime"}]], "Input"],

Cell[BoxData[
 FractionBox["6587.630109211179`", "Second"]], "Output",
 CellChangeTimes->{3.5086614942700577`*^9, 3.508661917608389*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " Compile"
}], "Subsubsection"],

Cell["Switch to the non external implementation", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DeactivateCode", "[", 
  RowBox[{"Ffunction", ",", " ", "FfunLoop"}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   TagBox["Ffunction",
    HoldForm], ",", 
   TagBox["FfunLoop",
    HoldForm]}], "}"}]], "Output",
 CellChangeTimes->{3.5086614942856827`*^9, 3.508661917624014*^9}]
}, Open  ]],

Cell["Use Compile on the Ffunction", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"fc", " ", "=", " ", 
   RowBox[{"Compile", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"EA", ",", " ", "_Real"}], "}"}], ",", "\n", "\t\t\t", 
       RowBox[{"{", 
        RowBox[{"EIx", ",", "_Real"}], "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"EIy", ",", " ", "_Real"}], "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"EIz", ",", " ", "_Real"}], "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"GAy", ",", " ", "_Real"}], "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"GAz", ",", " ", "_Real"}], "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"l", ",", " ", "_Real"}], "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"t1i", ",", " ", "_Real"}], "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"t1j", ",", " ", "_Real"}], "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"t2i", ",", " ", "_Real"}], "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"t2j", ",", " ", "_Real"}], "}"}], ",", 
       RowBox[{"{", " ", 
        RowBox[{"t3i", ",", " ", "_Real"}], "}"}], ",", 
       RowBox[{"{", " ", 
        RowBox[{"t3j", ",", " ", "_Real"}], "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"u1i", ",", " ", "_Real"}], "}"}], ",", 
       RowBox[{"{", " ", 
        RowBox[{"u1j", ",", " ", "_Real"}], "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"u2i", ",", " ", "_Real"}], "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"u2j", ",", " ", "_Real"}], "}"}], ",", 
       RowBox[{"{", " ", 
        RowBox[{"u3i", ",", " ", "_Real"}], "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"u3j", ",", " ", "_Real"}], "}"}], ",", " ", 
       RowBox[{"{", 
        RowBox[{"x", ",", " ", "_Real"}], "}"}]}], "}"}], ",", " ", 
     RowBox[{"Evaluate", "[", 
      RowBox[{"Ffunction", "[", 
       RowBox[{
       "EA", ",", " ", "EIx", ",", " ", "EIy", ",", " ", "EIz", ",", " ", 
        "GAy", ",", " ", "GAz", ",", " ", "l", ",", " ", "t1i", ",", " ", 
        "t1j", ",", " ", "t2i", ",", " ", "t2j", ",", " ", "t3i", ",", " ", 
        "t3j", ",", " ", "u1i", ",", " ", "u1j", ",", " ", "u2i", ",", " ", 
        "u2j", ",", " ", "u3i", ",", " ", "u3j", ",", " ", "x"}], "]"}], 
      "]"}]}], "]"}]}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{"(*", 
  RowBox[{
   RowBox[{"fc", "[", 
    RowBox[{"Sequence", "@@", " ", 
     RowBox[{"Table", "[", 
      RowBox[{"0.65", ",", 
       RowBox[{"{", "20", "}"}]}], "]"}]}], "]"}], " ", "//", " ", "Timing"}],
   "*)"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"FcfunLoop", "[", 
      RowBox[{"Integer", " ", "n_"}], "]"}], "->", 
     RowBox[{"Real", "[", "_", "]"}]}], " ", ":=", " ", "\n", "\t", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"Real", "[", "_", "]"}], "  ", "res"}], "}"}], ",", "\n", 
      "\t\t", 
      RowBox[{
       RowBox[{"Print", "[", "n", "]"}], ";", "\n", "\t\t", 
       RowBox[{"Do", "[", 
        RowBox[{
         RowBox[{"res", " ", "=", 
          RowBox[{"fc", "[", 
           RowBox[{"Sequence", "@@", " ", 
            RowBox[{"Table", "[", 
             RowBox[{"0.65", ",", 
              RowBox[{"{", "20", "}"}]}], "]"}]}], "]"}]}], ",", 
         RowBox[{"{", "n", "}"}]}], "]"}], ";", "\n", "\t\t", "res"}]}], "\n",
      "\t\t", "]"}]}], ";"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"General", "::", "spell1"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"New symbol name \\\"\[NoBreak]\\!\\(FcfunLoop\\)\[NoBreak]\
\\\" is similar to existing symbol \
\\\"\[NoBreak]\\!\\(FfunLoop\\)\[NoBreak]\\\" and may be misspelled. \
\\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\\\", \
ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/General/spell1\\\", \
ButtonNote -> \\\"General::spell1\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.508661917686513*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"FcfunLoop", "[", "2000", "]"}], "//", "Timing"}], 
  "\[IndentingNewLine]"}]], "Input"],

Cell[BoxData["2000"], "Print",
 CellChangeTimes->{3.5086614943638067`*^9, 3.508661917702138*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.03100000000000011`", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.3797168294052363`"}], ",", "0.7109900453920844`", ",", 
     RowBox[{"-", "0.33127321598684806`"}], ",", "0.`"}], "}"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{3.508661494395056*^9, 3.508661917733387*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"comptime", " ", "=", " ", 
   RowBox[{
    RowBox[{"%", "[", 
     RowBox[{"[", "1", "]"}], "]"}], "/", "2000"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"mathtime", " ", "/", " ", "comptime"}], " ", 
  RowBox[{"(*", 
   RowBox[{"Efficiency", " ", "of", " ", 
    RowBox[{"Mathematica", "'"}], "s", " ", 
    RowBox[{"Compile", "[", "]"}]}], "*)"}]}]], "Input"],

Cell[BoxData["604.3010752688144`"], "Output",
 CellChangeTimes->{3.508661494426306*^9, 3.508661917749012*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"comptime", " ", "/", " ", "exttime"}], " ", 
  RowBox[{"(*", 
   RowBox[{"Efficiency", " ", "of", " ", "MathCode", " ", "over", " ", 
    RowBox[{"Mathematica", "'"}], "s", " ", 
    RowBox[{"Compile", "[", "]"}]}], "*)"}]}]], "Input"],

Cell[BoxData[
 FractionBox["10.901238436950933`", "Second"]], "Output",
 CellChangeTimes->{3.5086614944419303`*^9, 3.5086619177802615`*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}]], "Input"],

Cell["CleanMathCodeFiles[Confirm -> False, CleanAllBut -> {}];", "Input"],

Cell["\<\
Even more performance improvement can be obtained for code including loops\
\>", "Itemize"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
ScreenStyleEnvironment->"Working",
WindowToolbars->{},
WindowSize->{678, 910},
WindowMargins->{{8, Automatic}, {Automatic, 5}},
ShowCellTags->True,
FrontEndVersion->"6.0 for Microsoft Windows (32-bit) (April 20, 2007)",
StyleDefinitions->Notebook[{
   Cell[
    CellGroupData[{
      Cell["Style Definitions", "Subtitle"], 
      Cell[
      "\nModify the definitions below to change the default appearance of all \
cells in \na given style.  Make modifications to any definition using \
commands in the \nFormat menu.\n", "Text"], 
      Cell[
       CellGroupData[{
         Cell["Style Environment Names", "Section"], 
         Cell[
          StyleData[All, "Working"], PageWidth -> WindowWidth, ScriptMinSize -> 
          9], 
         Cell[
          StyleData[All, "Presentation"], PageWidth -> WindowWidth, 
          ScriptMinSize -> 12, FontSize -> 16], 
         Cell[
          StyleData[All, "Condensed"], PageWidth -> WindowWidth, 
          CellBracketOptions -> {"Margins" -> {1, 1}, "Widths" -> {0, 5}}, 
          ScriptMinSize -> 8, FontSize -> 11], 
         Cell[
          StyleData[All, "Printout"], PageWidth -> PaperWidth, ScriptMinSize -> 
          7, FontSize -> 10, 
          PrivateFontOptions -> {"FontType" -> "Outline"}]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Notebook Options", "Section"], 
         Cell[
         "\nThe options defined for the style below will be used at the \
Notebook level.\n", "Text"], 
         Cell[
          StyleData["Notebook"], PageHeaders -> {{
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"], None, 
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"]}, {
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"], None, 
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"]}}, 
          PageHeaderLines -> {True, True}, 
          PrintingOptions -> {
           "FacingPages" -> True, "FirstPageHeader" -> False}, 
          CellFrameLabelMargins -> 6, StyleMenuListing -> None]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headings", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 0}, CounterIncrements -> "Title", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subtitle", 0}, {"Subsubtitle", 0}}, FontSize -> 
             34, FontColor -> GrayLevel[1], Background -> RGBColor[0, 0, 1]], 
            
            Cell[
             StyleData["Title", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 0}, FontSize -> 
             44], 
            Cell[
             StyleData["Title", "Condensed"], CellMargins -> {{0, 0}, {0, 0}},
              FontSize -> 20], 
            Cell[
             StyleData["Title", "Printout"], CellMargins -> {{0, 0}, {0, 0}}, 
             FontSize -> 24, FontTracking -> "Plain", Background -> 
             GrayLevel[0]]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 3}, ParagraphIndent -> -96, CounterIncrements -> 
             "Subtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subsubtitle", 0}}, FontFamily -> "Helvetica", 
             FontSize -> 18, FontColor -> GrayLevel[1], Background -> 
             RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Subtitle", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 6}, 
             ParagraphIndent -> -157, FontSize -> 30], 
            Cell[
             StyleData["Subtitle", "Condensed"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -78, 
             FontSize -> 14], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -85, 
             FontSize -> 16, Background -> GrayLevel[0.6]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {30, 10}}, PageBreakBelow -> False, 
             CounterIncrements -> "Subsubtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}}, FontFamily -> "Helvetica", FontSize -> 14, 
             FontSlant -> "Italic"], 
            Cell[
             StyleData["Subsubtitle", "Presentation"], 
             CellMargins -> {{8, 10}, {40, 20}}, LineSpacing -> {1, 0}, 
             FontSize -> 24], 
            Cell[
             StyleData["Subsubtitle", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 8}}, FontSize -> 12], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{9, 10}, {50, 10}}, FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], CellFrame -> {{6, 0}, {0, 1}}, CellDingbat -> 
             None, CellMargins -> {{12, Inherited}, {4, 24}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameMargins -> 6, CounterIncrements -> "Section", 
             CounterAssignments -> {{"Subsection", 0}, {"Subsubsection", 0}}, 
             FontFamily -> "Helvetica", FontSize -> 16, FontWeight -> "Bold", 
             FontColor -> RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Section", "Presentation"], 
             CellMargins -> {{10, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["Section", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Section", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, FontSize -> 14, FontTracking -> 
             "Plain", FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], 
             CellMargins -> {{12, Inherited}, {8, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 40}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsection", "Presentation"], 
             CellMargins -> {{11, 10}, {8, 32}}, LineSpacing -> {1, 0}, 
             FontSize -> 22], 
            Cell[
             StyleData["Subsection", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Subsection", "Printout"], 
             CellMargins -> {{9, 0}, {4, 40}}, FontSize -> 12]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> "\[FilledSquare]", 
             CellMargins -> {{25, Inherited}, {8, 12}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsubsection", "Presentation"], 
             CellMargins -> {{29, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Subsubsection", "Condensed"], 
             CellMargins -> {{22, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Subsubsection", "Printout"], 
             CellMargins -> {{21, 0}, {4, 20}}, FontSize -> 11]}, Open]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Body Text", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Text"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Text", FontFamily -> "Times"], 
            Cell[
             StyleData["Text", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Text", "Condensed"], CellMargins -> {{8, 10}, {4, 4}},
              LineSpacing -> {1, 1}, ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Text", "Printout"], CellMargins -> {{9, 0}, {4, 4}}, 
             ParagraphSpacing -> {0, 6}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SmallText"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 6}, 
             CounterIncrements -> "SmallText", FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["SmallText", "Condensed"], 
             CellMargins -> {{8, 10}, {2, 2}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Printout"], 
             CellMargins -> {{9, 0}, {4, 4}}, FontSize -> 7]}, Closed]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Input/Output", "Section"], 
         Cell[
         "\nThe cells in this section define styles used for input and output \
to the \nkernel.  Be careful when modifying, renaming, or removing these \
styles, \nbecause the front end associates special meanings with these style \
names.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Input"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             CounterIncrements -> "Input", FontWeight -> "Bold", Background -> 
             RGBColor[1, 1, 0]], 
            Cell[
             StyleData["Input", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Input", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["Input", "Printout"], CellMargins -> {{44, 0}, {6, 6}},
              Background -> GrayLevel[0.8]]}, Closed]], 
         Cell[
          StyleData["InlineInput"], Evaluatable -> True, CellGroupingRules -> 
          "InputGrouping", CellHorizontalScrolling -> True, PageBreakWithin -> 
          False, GroupPageBreakWithin -> False, DefaultFormatType -> 
          DefaultInputFormatType, AutoItalicWords -> {}, FormatType -> 
          InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
          CounterIncrements -> "Input", FontWeight -> "Bold"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Output"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, CellEditDuplicate -> True, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, 
             CellLabelMargins -> {{3, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Output", Background -> 
             RGBColor[1, 0.900008, 0.900008]], 
            Cell[
             StyleData["Output", "Presentation"], 
             CellMargins -> {{62, Inherited}, {12, 5}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Output", "Condensed"], 
             CellMargins -> {{40, Inherited}, {4, 1}}], 
            Cell[
             StyleData["Output", "Printout"], 
             CellMargins -> {{44, 0}, {6, 2}}, Background -> 
             GrayLevel[0.900008]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Message"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Message", StyleMenuListing -> 
             None, FontColor -> RGBColor[1, 0, 0]], 
            Cell[
             StyleData["Message", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Message", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Message", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Print"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, ShowCellLabel -> 
             False, DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Print", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Print", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Print", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Print", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Graphics"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "GraphicsGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Graphics", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Graphics", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Graphics", "Condensed"], 
             CellMargins -> {{52, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {175, 175}], 
            Cell[
             StyleData["Graphics", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {250, 250}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["CellLabel"], StyleMenuListing -> None, FontFamily -> 
             "Helvetica", FontSize -> 11, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["CellLabel", "Presentation"], FontSize -> 12], 
            Cell[
             StyleData["CellLabel", "Condensed"], FontSize -> 8], 
            Cell[
             StyleData["CellLabel", "Printout"], FontSize -> 8, FontColor -> 
             GrayLevel[0]]}, Closed]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Unique Styles", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {2, 10}}, FontSize -> 16, FontSlant -> 
             "Italic"], 
            Cell[
             StyleData["Author", "Presentation"], 
             CellMargins -> {{12, 10}, {2, 12}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}, FontSize -> 20], 
            Cell[
             StyleData["Author", "Condensed"], 
             CellMargins -> {{8, 10}, {1, 4}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}, FontSize -> 12], 
            Cell[
             StyleData["Author", "Printout"], 
             CellMargins -> {{9, 0}, {4, 12}}, ParagraphSpacing -> {0, 6}, 
             FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Copyright"], ShowCellBracket -> False, 
             CellMargins -> {{10, 10}, {40, 2}}, FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Presentation"], 
             CellMargins -> {{12, 10}, {50, 2}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["Copyright", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 1}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Printout"], 
             CellMargins -> {{9, 0}, {72, 4}}, FontSize -> 7]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Example"], CellMargins -> {{12, 10}, {5, 12}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Example", FontFamily -> "Times", 
             FontWeight -> "Bold"], 
            Cell[
             StyleData["Example", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 20}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Example", "Condensed"], 
             CellMargins -> {{8, 10}, {4, 8}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Example", "Printout"], 
             CellMargins -> {{9, 0}, {4, 10}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Definition"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Definition", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Definition", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Definition", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Definition", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.6]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Theorem"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Theorem", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Theorem", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Theorem", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Theorem", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.4]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseMain"], CellFrame -> {{6, 0}, {0, 1}}, 
             CellMargins -> {{12, Inherited}, {4, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameColor -> RGBColor[0.571389, 0.19675, 0.570504], 
             CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 16, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["ExerciseMain", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["ExerciseMain", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["ExerciseMain", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, CellFrameColor -> 
             GrayLevel[0.500008], FontSize -> 14, FontTracking -> "Plain", 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Exercise"], CellDingbat -> "\[FilledDownTriangle]", 
             CellMargins -> {{23, Inherited}, {4, 18}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Exercise", "Presentation"], 
             CellMargins -> {{33, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Exercise", "Condensed"], 
             CellMargins -> {{17, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Exercise", "Printout"], 
             CellFrame -> {{0, 0}, {0.5, 0}}, CellDingbat -> None, 
             CellMargins -> {{9, 0}, {6, 20}}, FontSize -> 11, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseText"], CellMargins -> {{62, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 8}, 
             CounterIncrements -> "ExcersiceText", FontFamily -> "Times"], 
            Cell[
             StyleData["ExerciseText", "Presentation"], 
             CellMargins -> {{74, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["ExerciseText", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 2}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["ExerciseText", "Printout"], 
             CellMargins -> {{54, 0}, {4, 4}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Itemize"], CellDingbat -> "\[FilledSmallCircle]", 
             CellMargins -> {{65, Inherited}, {8, 8}}, FontFamily -> "Times", 
             FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Itemize", "Presentation"], 
             CellMargins -> {{Inherited, 10}, {8, 12}}, LineSpacing -> {1, 0},
              FontSize -> 18], 
            Cell[
             StyleData["Itemize", "Condensed"], 
             CellMargins -> {{Inherited, Inherited}, {2, 12}}, FontSize -> 
             10], 
            Cell[
             StyleData["Itemize", "Printout"], 
             CellMargins -> {{Inherited, 0}, {4, 20}}, FontSize -> 11]}, 
           Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayCode"], CellFrame -> 2, 
             CellMargins -> {{52, 10}, {8, 8}}, CellHorizontalScrolling -> 
             True, PageBreakWithin -> True, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             FontWeight -> "Bold", Background -> RGBColor[1, 1, 0]], 
            Cell[
             StyleData["DisplayCode", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["DisplayCode", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["DisplayCode", "Printout"], 
             CellMargins -> {{44, 0}, {6, 6}}, Background -> GrayLevel[0.8]]},
            Closed]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Formulas and Programming", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayFormula"], CellMargins -> {{62, 10}, {2, 10}}, 
             CellHorizontalScrolling -> True, ScriptLevel -> 0, 
             SingleLetterItalics -> True, 
             UnderoverscriptBoxOptions -> {LimitsPositioning -> True}], 
            Cell[
             StyleData["DisplayFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 18], 
            Cell[
             StyleData["DisplayFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["DisplayFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ChemicalFormula"], CellMargins -> {{62, 10}, {2, 10}},
              AutoSpacing -> False, ScriptBaselineShifts -> {0.6, Automatic}, 
             ScriptLevel -> 1, SingleLetterItalics -> False, ZeroWidthTimes -> 
             True], 
            Cell[
             StyleData["ChemicalFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Program"], CellMargins -> {{12, 10}, {Inherited, 6}}, 
             FontFamily -> "Courier"], 
            Cell[
             StyleData["Program", "Presentation"], 
             CellMargins -> {{13, 30}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Condensed"], 
             CellMargins -> {{8, 10}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Printout"], 
             CellMargins -> {{9, 0}, {Inherited, 4}}, FontSize -> 9.5]}, 
           Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Automatic Numbering", "Section"], 
         Cell[
         "\nThe following styles are useful for numbered equations, figures, \
etc.  They \nautomatically give the cell a FrameLabel containing a reference \
to a \nparticular counter, and also increment that counter.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedEquation"], 
             CellMargins -> {{62, 10}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, 
                Cell[
                 TextData[{"(", 
                   CounterBox["NumberedEquation"], ")"}]]}, {None, None}}, 
             DefaultFormatType -> DefaultInputFormatType, CounterIncrements -> 
             "NumberedEquation", FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedEquation", "Presentation"], 
             CellMargins -> {{74, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Condensed"], 
             CellMargins -> {{52, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Printout"], 
             CellMargins -> {{54, 0}, {Inherited, Inherited}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedFigure"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Figure ", 
                   CounterBox["NumberedFigure"]}]], None}}, CounterIncrements -> 
             "NumberedFigure", 
             ImageMargins -> {{43, Inherited}, {Inherited, 0}}, 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedFigure", "Presentation"]], 
            Cell[
             StyleData["NumberedFigure", "Condensed"]], 
            Cell[
             StyleData["NumberedFigure", "Printout"]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedTable"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Table ", 
                   CounterBox["NumberedTable"]}]], None}}, TextAlignment -> 
             Center, CounterIncrements -> "NumberedTable", 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedTable", "Presentation"]], 
            Cell[
             StyleData["NumberedTable", "Condensed"]], 
            Cell[
             StyleData["NumberedTable", "Printout"]]}, Open]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headers and Footers", "Section"], 
         Cell[
          StyleData["Header"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontSlant -> "Italic"], 
         Cell[
          StyleData["Footer"], CellMargins -> {{0, 0}, {0, 4}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 6],
          
         Cell[
          StyleData["PageNumber"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontWeight -> "Bold"]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Hyperlink Styles", "Section"], 
         Cell[
         "\nThe cells below define styles useful for making hypertext \
ButtonBoxes.  The \n\"Hyperlink\" style is for links within the same \
Notebook, or between \nNotebooks.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Hyperlink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, 
             FontVariations -> {"Underline" -> True}, FontColor -> 
             RGBColor[0, 0, 1], Background -> None, 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`NotebookLocate[#2]}]& ), ButtonNote -> 
               ButtonData}], 
            Cell[
             StyleData["Hyperlink", "Presentation"]], 
            Cell[
             StyleData["Hyperlink", "Condensed"]], 
            Cell[
             StyleData["Hyperlink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Closed]], 
         Cell[
         "\nThe following styles are for linking automatically to the on-line \
help \nsystem.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["MainBookLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["MainBook", #]}]& )}], 
            Cell[
             StyleData["MainBookLink", "Presentation"]], 
            Cell[
             StyleData["MainBookLink", "Condensed"]], 
            Cell[
             StyleData["MainBookLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["AddOnsLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["AddOns", #]}]& )}], 
            Cell[
             StyleData["AddOnsLink", "Presentation"]], 
            Cell[
             StyleData["AddOnsLink", "Condensed"]], 
            Cell[
             StyleData["AddOnLink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["RefGuideLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["RefGuideLink", #]}]& )}], 
            Cell[
             StyleData["RefGuideLink", "Presentation"]], 
            Cell[
             StyleData["RefGuideLink", "Condensed"]], 
            Cell[
             StyleData["RefGuideLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["GettingStartedLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["GettingStarted", #]}]& )}], 
            Cell[
             StyleData["GettingStartedLink", "Presentation"]], 
            Cell[
             StyleData["GettingStartedLink", "Condensed"]], 
            Cell[
             StyleData["GettingStartedLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["OtherInformationLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["OtherInformation", #]}]& )}], 
            Cell[
             StyleData["OtherInformationLink", "Presentation"]], 
            Cell[
             StyleData["OtherInformationLink", "Condensed"]], 
            Cell[
             StyleData["OtherInformationLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]]}, Open]]}, 
     Open]]}, Visible -> False, FrontEndVersion -> 
  "6.0 for Microsoft Windows (32-bit) (April 20, 2007)", StyleDefinitions -> 
  "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[568, 21, 33, 0, 55, "Title"],
Cell[604, 23, 58, 0, 37, "Subtitle"],
Cell[665, 25, 73, 2, 54, "Copyright"],
Cell[CellGroupData[{
Cell[763, 31, 44, 0, 54, "Section"],
Cell[810, 33, 117, 3, 25, "Text"],
Cell[930, 38, 57, 0, 31, "Itemize"],
Cell[990, 40, 43, 0, 31, "Itemize"],
Cell[CellGroupData[{
Cell[1058, 44, 49, 0, 44, "Subsection"],
Cell[1110, 46, 379, 13, 43, "Text"],
Cell[1492, 61, 110, 3, 48, "Input"],
Cell[CellGroupData[{
Cell[1627, 68, 46, 0, 35, "Subsubsection"],
Cell[1676, 70, 6710, 210, 1149, "Input"],
Cell[CellGroupData[{
Cell[8411, 284, 70, 1, 48, "Input"],
Cell[8484, 287, 1236, 38, 117, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[9769, 331, 51, 0, 35, "Subsubsection"],
Cell[9823, 333, 630, 15, 129, "Input"],
Cell[10456, 350, 644, 18, 109, "Input"],
Cell[11103, 370, 64, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[11192, 374, 62, 1, 48, "Input"],
Cell[11257, 377, 102, 1, 47, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[11420, 385, 35, 0, 44, "Subsection"],
Cell[11458, 387, 61, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[11544, 391, 104, 5, 48, "Input"],
Cell[11651, 398, 450, 10, 23, "Print"]
}, Open  ]],
Cell[12116, 411, 66, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[12207, 415, 101, 2, 48, "Input"],
Cell[12311, 419, 106, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[12454, 425, 296, 9, 167, "Input"],
Cell[12753, 436, 551, 11, 21, "Message"]
}, Open  ]],
Cell[13319, 450, 107, 3, 25, "Text"],
Cell[13429, 455, 67, 0, 48, "Input"],
Cell[13499, 457, 1170, 28, 269, "Input"],
Cell[14672, 487, 170, 3, 43, "Text"],
Cell[CellGroupData[{
Cell[14867, 494, 24, 0, 48, "Input"],
Cell[14894, 496, 102, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[15033, 502, 134, 3, 48, "Input"],
Cell[15170, 507, 146, 1, 23, "Print"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[15365, 514, 43, 0, 44, "Subsection"],
Cell[15411, 516, 62, 0, 25, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[15510, 521, 41, 0, 44, "Subsection"],
Cell[15554, 523, 35, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[15614, 527, 53, 1, 48, "Input"],
Cell[15670, 530, 100, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[15807, 536, 112, 5, 36, "Subsubsection"],
Cell[CellGroupData[{
Cell[15944, 545, 232, 7, 48, "Input"],
Cell[16179, 554, 314, 8, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[16530, 567, 118, 3, 48, "Input"],
Cell[16651, 572, 96, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[16784, 578, 128, 3, 69, "Input"],
Cell[16915, 583, 94, 1, 23, "Print"],
Cell[17012, 586, 330, 8, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[17379, 599, 183, 5, 69, "Input"],
Cell[17565, 606, 114, 1, 47, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[17728, 613, 45, 0, 35, "Subsubsection"],
Cell[17776, 615, 48, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[17849, 619, 77, 2, 48, "Input"],
Cell[17929, 623, 248, 5, 23, "Print"]
}, Open  ]],
Cell[18192, 631, 44, 0, 25, "Text"],
Cell[18239, 633, 783, 22, 209, "Input"],
Cell[19025, 657, 135, 5, 25, "Text"],
Cell[CellGroupData[{
Cell[19185, 666, 101, 2, 48, "Input"],
Cell[19289, 670, 366, 9, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[19692, 684, 158, 4, 48, "Input"],
Cell[19853, 690, 169, 3, 47, "Output"]
}, Open  ]],
Cell[20037, 696, 32, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[20094, 700, 72, 1, 48, "Input"],
Cell[20169, 703, 136, 2, 64, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[20354, 711, 97, 4, 36, "Subsubsection"],
Cell[20454, 717, 57, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[20536, 721, 110, 2, 48, "Input"],
Cell[20649, 725, 206, 7, 47, "Output"]
}, Open  ]],
Cell[20870, 735, 44, 0, 25, "Text"],
Cell[20917, 737, 2299, 55, 169, "Input"],
Cell[23219, 794, 255, 8, 48, "Input"],
Cell[CellGroupData[{
Cell[23499, 806, 888, 25, 169, "Input"],
Cell[24390, 833, 551, 11, 21, "Message"]
}, Open  ]],
Cell[CellGroupData[{
Cell[24978, 849, 134, 4, 69, "Input"],
Cell[25115, 855, 96, 1, 23, "Print"],
Cell[25214, 858, 327, 8, 47, "Output"]
}, Open  ]],
Cell[25556, 869, 167, 5, 48, "Input"],
Cell[CellGroupData[{
Cell[25748, 878, 241, 6, 48, "Input"],
Cell[25992, 886, 109, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[26138, 892, 270, 6, 48, "Input"],
Cell[26411, 900, 139, 2, 64, "Output"]
}, Open  ]],
Cell[26565, 905, 79, 2, 48, "Input"],
Cell[26647, 909, 73, 0, 48, "Input"],
Cell[26723, 911, 101, 2, 31, "Itemize"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
