(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     66634,       1746]
NotebookOptionsPosition[     25829,        817]
NotebookOutlinePosition[     63099,       1630]
CellTagsIndexPosition[     63056,       1627]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{
Cell["Replacement for NIntegrate", "Title"],

Cell[TextData[{
 StyleBox["(c) MathCore, 1998-2004\n(The whole notebook can be evaluated et \
once)",
  FontWeight->"Bold"],
 "\n"
}], "Text"],

Cell["Trapezoidal method", "Subtitle"],

Cell["Assume you want to integrate this function", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"f", "[", "x_", "]"}], ":=", 
  RowBox[{"1", "-", 
   RowBox[{"x", "^", "2"}], "+", 
   RowBox[{"2", "*", 
    RowBox[{"Sin", "[", "x", "]"}]}]}]}]], "Input"],

Cell["You would like to write another function, like this", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"fint", "[", 
   RowBox[{"a_", ",", "b_"}], "]"}], ":=", 
  RowBox[{"NIntegrate", "[", 
   RowBox[{
    RowBox[{"f", "[", "z", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"z", ",", "a", ",", "b"}], "}"}]}], "]"}]}]], "Input"],

Cell["And be able to evaluate ", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"fint", "[", 
  RowBox[{"0", ",", "64"}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"-", "87316.11704779424`"}]], "Output",
 CellChangeTimes->{3.5086619998104615`*^9}]
}, Open  ]],

Cell["\<\
The simplest method is to add one more function that computes the integral \
using the Trapezoidal method. Number of intervals (which is normally power of \
2) should be specified.\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"fintIntervals", "[", 
   RowBox[{"a_", ",", "b_", ",", "n_"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"s", ",", "i", ",", "j", ",", "dx"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"s", "=", 
      RowBox[{
       RowBox[{"f", "[", "a", "]"}], "/", "2"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"dx", "=", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"b", "-", "a"}], ")"}], "/", 
       RowBox[{"(", 
        RowBox[{"2", "^", "n"}], ")"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"For", "[", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", 
       RowBox[{"j", "<", 
        RowBox[{"2", "^", "n"}]}], ",", 
       RowBox[{"j", "+=", "1"}], ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"i", "=", 
         RowBox[{"a", "+", 
          RowBox[{"j", "*", "dx"}]}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"s", "+=", 
         RowBox[{"f", "[", "i", "]"}]}]}]}], "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"s", "+=", 
      RowBox[{
       RowBox[{"f", "[", "b", "]"}], "/", "2"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"s", "*", "dx"}]}]}], "\[IndentingNewLine]", "]"}]}]], "Input"],

Cell["\<\

Assuming you guess a good number of intervals, for instance 2^9 =512 (see \
later - how to do that), and get appoximately same result:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"fintIntervals", "[", 
  RowBox[{"0.", ",", "64.", ",", "9"}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"-", "87316.28529857853`"}]], "Output",
 CellChangeTimes->{3.508661999857336*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{"fintIntervals", "[", 
     RowBox[{"0.", ",", "64.", ",", "9"}], "]"}], " ", "-", " ", 
    RowBox[{"fint", "[", 
     RowBox[{"0", ",", "64"}], "]"}]}], ")"}], "/", 
  RowBox[{"fint", "[", 
   RowBox[{"0", ",", "64"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.5086620744970055`*^9, 3.508662075481368*^9}}],

Cell[BoxData["1.926915556727888`*^-6"], "Output",
 CellChangeTimes->{3.50866199990421*^9}]
}, Open  ]],

Cell["Running this with MathCode", "Subtitle"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Needs", "[", "\"\<MathCode`\>\"", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"MathCode C++ 1.4.4\"\>", "\[InvisibleSpace]", "\<\" for \"\>", 
   "\[InvisibleSpace]", "\<\"vc100\"\>", 
   "\[InvisibleSpace]", "\<\" loaded from \"\>", 
   "\[InvisibleSpace]", "\<\"C:\\\\b\\\\omath\\\\MathCode\"\>"}],
  SequenceForm[
  "MathCode C++ 1.4.4", " for ", "vc100", " loaded from ", 
   "C:\\b\\omath\\MathCode"],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086620002479563`*^9}]
}, Open  ]],

Cell["\<\
SetDirectory[$MCRoot<>$PathnameSeparator<>\"Demos\"<>$PathnameSeparator<>\"\
NIntegrate\"];\
\>", "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"f", "[", 
     RowBox[{"Real", " ", "x_"}], "]"}], "\[Rule]", "Real"}], ":=", 
   RowBox[{"1", "-", 
    RowBox[{"x", "^", "2"}], "+", 
    RowBox[{"2", "*", 
     RowBox[{"Sin", "[", "x", "]"}]}]}]}], "\[IndentingNewLine]"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"fintIntervals", "[", 
    RowBox[{
     RowBox[{"Real", " ", "a_"}], ",", 
     RowBox[{"Real", " ", "b_"}], ",", 
     RowBox[{"Integer", " ", "n_"}]}], "]"}], "\[Rule]", "Real"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"Real", " ", "s"}], ",", 
      RowBox[{"Real", " ", "i"}], ",", 
      RowBox[{"Real", " ", "dx"}], ",", 
      RowBox[{"Integer", " ", "j"}]}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"s", "=", 
      RowBox[{
       RowBox[{"f", "[", "a", "]"}], "/", "2"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"dx", "=", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"b", "-", "a"}], ")"}], "/", 
       RowBox[{"(", 
        RowBox[{"2", "^", "n"}], ")"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"For", "[", 
      RowBox[{
       RowBox[{"j", "=", "1"}], ",", 
       RowBox[{"j", "<", 
        RowBox[{"2", "^", "n"}]}], ",", 
       RowBox[{"j", "+=", "1"}], ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"i", "=", 
         RowBox[{"a", "+", 
          RowBox[{"j", "*", "dx"}]}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"s", "+=", 
         RowBox[{"f", "[", "i", "]"}]}]}]}], "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"s", "+=", 
      RowBox[{
       RowBox[{"f", "[", "b", "]"}], "/", "2"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"s", "*", "dx"}]}]}], "\[IndentingNewLine]", "]"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"BuildCode", "[", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "fint", "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`fint}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086620005292025`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "fint", "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`fint}],
  Editable->False]], "Print",
 CellChangeTimes->{3.508662000560452*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "fint", "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`fint}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086620005760765`*^9}],

Cell[BoxData["\<\"Successful compilation to C++: 2 function(s)\"\>"], "Print",
 CellChangeTimes->{3.508662002747924*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"InstallCode", "[", "]"}], ";"}]}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.508662006560375*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"fintIntervals", "[", 
   RowBox[{"0.", ",", "64.", ",", "9"}], "]"}], 
  "\[IndentingNewLine]"}]], "Input"],

Cell[BoxData[
 RowBox[{"-", "87316.28529857853`"}]], "Output",
 CellChangeTimes->{3.5086620066072493`*^9}]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"CleanMathCodeFiles", "[", 
   RowBox[{
    RowBox[{"Confirm", "\[Rule]", "False"}], ",", 
    RowBox[{"CleanAllBut", "\[Rule]", 
     RowBox[{"{", "}"}]}]}], "]"}], ";"}]}], "Input"],

Cell["Finding good number of intervals", "Subtitle"],

Cell[TextData[{
 "\nFinding good number of intervals requires some experimentation.    First \
you have to decide the precision you require. Standard value for ",
 StyleBox["PrecisionGoal",
  FontWeight->"Bold"],
 " option of ",
 StyleBox["NIntegral",
  FontWeight->"Bold"],
 " is",
 StyleBox[" 8",
  FontWeight->"Bold"],
 ". Assume that ",
 StyleBox["5 ",
  FontWeight->"Bold"],
 "digits is sufficient for your goals.     Evaluate your integral with the \
options below:\n"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NIntegrate", " ", "[", 
  RowBox[{
   RowBox[{"f", "[", "x", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "0.", ",", "64."}], "}"}], ",", 
   RowBox[{"Method", "\[Rule]", " ", "Trapezoidal"}], ",", 
   RowBox[{"PrecisionGoal", "\[Rule]", "5"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"-", "87316.28529857859`"}]], "Output",
 CellChangeTimes->{3.5086620077166104`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NIntegrate", " ", "[", 
  RowBox[{
   RowBox[{"f", "[", "x", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "0.", ",", "64."}], "}"}], ",", 
   RowBox[{"Method", "\[Rule]", " ", "Trapezoidal"}], ",", 
   RowBox[{"MaxRecursion", "\[Rule]", "1"}], ",", 
   RowBox[{"PrecisionGoal", "\[Rule]", "5"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"NIntegrate", "::", "ncvi"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"NIntegrate failed to converge to prescribed accuracy after \
\[NoBreak]\\!\\(1\\)\[NoBreak] iterated refinements in \[NoBreak]\\!\\(x\\)\
\[NoBreak] in the region \[NoBreak]\\!\\({\\({0.`, 64.`}\\)}\\)\[NoBreak]. \
NIntegrate obtained \[NoBreak]\\!\\(-87489.1132845305`\\)\[NoBreak] and \
\[NoBreak]\\!\\(506.68473522373824`\\)\[NoBreak] for the integral and error \
estimates. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\
\\\", ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/NIntegrate/ncvi\\\
\", ButtonNote -> \\\"NIntegrate::ncvi\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.508662007825984*^9}],

Cell[BoxData[
 RowBox[{"-", "87489.1132845305`"}]], "Output",
 CellChangeTimes->{3.508662007825984*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NIntegrate", " ", "[", 
  RowBox[{
   RowBox[{"f", "[", "x", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "0.", ",", "64."}], "}"}], ",", 
   RowBox[{"Method", "\[Rule]", " ", "Trapezoidal"}], ",", 
   RowBox[{"MaxRecursion", "\[Rule]", "2"}], ",", 
   RowBox[{"PrecisionGoal", "\[Rule]", "5"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"NIntegrate", "::", "ncvi"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"NIntegrate failed to converge to prescribed accuracy after \
\[NoBreak]\\!\\(2\\)\[NoBreak] iterated refinements in \[NoBreak]\\!\\(x\\)\
\[NoBreak] in the region \[NoBreak]\\!\\({\\({0.`, 64.`}\\)}\\)\[NoBreak]. \
NIntegrate obtained \[NoBreak]\\!\\(-87359.21903203645`\\)\[NoBreak] and \
\[NoBreak]\\!\\(129.8942524940503`\\)\[NoBreak] for the integral and error \
estimates. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\
\\\", ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/NIntegrate/ncvi\\\
\", ButtonNote -> \\\"NIntegrate::ncvi\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.508662007841609*^9}],

Cell[BoxData[
 RowBox[{"-", "87359.21903203645`"}]], "Output",
 CellChangeTimes->{3.508662007841609*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{"NIntegrate", " ", "[", 
   RowBox[{
    RowBox[{"f", "[", "x", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"x", ",", "0.", ",", "64."}], "}"}], ",", 
    RowBox[{"Method", "\[Rule]", " ", "Trapezoidal"}], ",", 
    RowBox[{"MaxRecursion", "\[Rule]", "3"}], ",", 
    RowBox[{"PrecisionGoal", "\[Rule]", "5"}]}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"NIntegrate", "::", "ncvi"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"NIntegrate failed to converge to prescribed accuracy after \
\[NoBreak]\\!\\(3\\)\[NoBreak] iterated refinements in \[NoBreak]\\!\\(x\\)\
\[NoBreak] in the region \[NoBreak]\\!\\({\\({0.`, 64.`}\\)}\\)\[NoBreak]. \
NIntegrate obtained \[NoBreak]\\!\\(-87326.88680212725`\\)\[NoBreak] and \
\[NoBreak]\\!\\(32.33222990919603`\\)\[NoBreak] for the integral and error \
estimates. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\
\\\", ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/NIntegrate/ncvi\\\
\", ButtonNote -> \\\"NIntegrate::ncvi\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.5086620078572335`*^9}],

Cell[BoxData[
 RowBox[{"-", "87326.88680212725`"}]], "Output",
 CellChangeTimes->{3.5086620078572335`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{"NIntegrate", " ", "[", 
   RowBox[{
    RowBox[{"f", "[", "x", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"x", ",", "0.", ",", "64."}], "}"}], ",", 
    RowBox[{"Method", "\[Rule]", " ", "Trapezoidal"}], ",", 
    RowBox[{"MaxRecursion", "\[Rule]", "4"}], ",", 
    RowBox[{"PrecisionGoal", "\[Rule]", "5"}]}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"NIntegrate", "::", "ncvi"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"NIntegrate failed to converge to prescribed accuracy after \
\[NoBreak]\\!\\(4\\)\[NoBreak] iterated refinements in \[NoBreak]\\!\\(x\\)\
\[NoBreak] in the region \[NoBreak]\\!\\({\\({0.`, 64.`}\\)}\\)\[NoBreak]. \
NIntegrate obtained \[NoBreak]\\!\\(-87318.80915995568`\\)\[NoBreak] and \
\[NoBreak]\\!\\(8.077642171570915`\\)\[NoBreak] for the integral and error \
estimates. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\
\\\", ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/NIntegrate/ncvi\\\
\", ButtonNote -> \\\"NIntegrate::ncvi\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.508662007872858*^9}],

Cell[BoxData[
 RowBox[{"-", "87318.80915995568`"}]], "Output",
 CellChangeTimes->{3.508662007872858*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{"NIntegrate", " ", "[", 
   RowBox[{
    RowBox[{"f", "[", "x", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"x", ",", "0.", ",", "64."}], "}"}], ",", 
    RowBox[{"Method", "\[Rule]", " ", "Trapezoidal"}], ",", 
    RowBox[{"MaxRecursion", "\[Rule]", "5"}], ",", 
    RowBox[{"PrecisionGoal", "\[Rule]", "5"}]}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"NIntegrate", "::", "ncvi"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"NIntegrate failed to converge to prescribed accuracy after \
\[NoBreak]\\!\\(5\\)\[NoBreak] iterated refinements in \[NoBreak]\\!\\(x\\)\
\[NoBreak] in the region \[NoBreak]\\!\\({\\({0.`, 64.`}\\)}\\)\[NoBreak]. \
NIntegrate obtained \[NoBreak]\\!\\(-87316.79005588999`\\)\[NoBreak] and \
\[NoBreak]\\!\\(2.019104065693682`\\)\[NoBreak] for the integral and error \
estimates. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\\\"Link\
\\\", ButtonFrame->None, ButtonData:>\\\"paclet:ref/message/NIntegrate/ncvi\\\
\", ButtonNote -> \\\"NIntegrate::ncvi\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.508662007888483*^9}],

Cell[BoxData[
 RowBox[{"-", "87316.79005588999`"}]], "Output",
 CellChangeTimes->{3.508662007888483*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{"NIntegrate", " ", "[", 
   RowBox[{
    RowBox[{"f", "[", "x", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"x", ",", "0.", ",", "64."}], "}"}], ",", 
    RowBox[{"Method", "\[Rule]", " ", "Trapezoidal"}], ",", 
    RowBox[{"MaxRecursion", "\[Rule]", "6"}], ",", 
    RowBox[{"PrecisionGoal", "\[Rule]", "5"}]}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"-", "87316.28529857859`"}]], "Output",
 CellChangeTimes->{3.508662007904108*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{"NIntegrate", " ", "[", 
   RowBox[{
    RowBox[{"f", "[", "x", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"x", ",", "0.", ",", "64."}], "}"}], ",", 
    RowBox[{"Method", "\[Rule]", " ", "Trapezoidal"}], ",", 
    RowBox[{"MaxRecursion", "\[Rule]", "7"}], ",", 
    RowBox[{"PrecisionGoal", "\[Rule]", "5"}]}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"-", "87316.28529857859`"}]], "Output",
 CellChangeTimes->{3.508662007919733*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{"NIntegrate", " ", "[", 
   RowBox[{
    RowBox[{"f", "[", "x", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"x", ",", "0.", ",", "64."}], "}"}], ",", 
    RowBox[{"Method", "\[Rule]", " ", "Trapezoidal"}], ",", 
    RowBox[{"MaxRecursion", "\[Rule]", "8"}], ",", 
    RowBox[{"PrecisionGoal", "\[Rule]", "5"}]}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"-", "87316.28529857859`"}]], "Output",
 CellChangeTimes->{3.5086620079353576`*^9}]
}, Open  ]],

Cell[TextData[{
 "You see that 6 recursions is sufficient for precision you need. We \
investigated how ",
 StyleBox["Trapeziodal",
  FontSlant->"Italic"],
 " method works and found out that ",
 StyleBox["2^(Recursion+3)",
  FontSlant->"Italic"],
 " intervals are used. Therefore the last parameter of ",
 StyleBox["fintIntervals",
  FontWeight->"Bold"],
 " is chosen ",
 StyleBox["9",
  FontWeight->"Bold"],
 ".\n"
}], "Text"],

Cell["Two Dimensional NIntegrate[], running with MathCode", "Subtitle"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"funt2a", "[", 
     RowBox[{
      RowBox[{"Real", " ", "a1_"}], ",", 
      RowBox[{"Real", " ", "b1_"}], ",", 
      RowBox[{"Integer", " ", "n1_"}], ",", "\[IndentingNewLine]", 
      RowBox[{"Real", " ", "a2_"}], ",", 
      RowBox[{"Real", " ", "b2_"}], ",", 
      RowBox[{"Integer", " ", "n2_"}]}], "]"}], "\[Rule]", "Real"}], ":=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"Real", " ", "s"}], ",", 
       RowBox[{"Integer", " ", "j1"}], ",", 
       RowBox[{"Integer", " ", "j2"}], ",", 
       RowBox[{"Real", " ", "dx1"}], ",", 
       RowBox[{"Real", " ", "dx2"}]}], "}"}], ",", "\[IndentingNewLine]", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"dx1", "=", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"b1", "-", "a1"}], ")"}], "/", 
        RowBox[{"(", 
         RowBox[{"2", "^", "n1"}], ")"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"dx2", "=", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{"b2", "-", "a2"}], ")"}], "/", 
        RowBox[{"(", 
         RowBox[{"2", "^", "n2"}], ")"}]}]}], ";", "\[IndentingNewLine]", 
      "\[IndentingNewLine]", 
      RowBox[{"s", "=", 
       RowBox[{
        RowBox[{"f2", "[", 
         RowBox[{"a1", ",", "a2"}], "]"}], "/", "4"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"For", " ", "[", 
       RowBox[{
        RowBox[{"j2", "=", "1"}], ",", 
        RowBox[{"j2", "<", 
         RowBox[{"2", "^", "n2"}]}], ",", 
        RowBox[{"j2", "+=", "1"}], ",", 
        RowBox[{"s", "+=", 
         RowBox[{
          RowBox[{"f2", "[", 
           RowBox[{"a1", ",", 
            RowBox[{"a2", "+", 
             RowBox[{"j2", "*", "dx2"}]}]}], "]"}], "/", "2"}]}]}], "]"}], 
      ";", "\[IndentingNewLine]", 
      RowBox[{"s", "+=", 
       RowBox[{
        RowBox[{"f2", "[", 
         RowBox[{"a1", ",", "b2"}], "]"}], "/", "4"}]}], ";", 
      "\[IndentingNewLine]", "\[IndentingNewLine]", 
      RowBox[{"For", " ", "[", 
       RowBox[{
        RowBox[{"j1", "=", "1"}], ",", 
        RowBox[{"j1", "<", 
         RowBox[{"2", "^", "n1"}]}], ",", 
        RowBox[{"j1", "+=", "1"}], ",", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"s", "+=", 
          RowBox[{
           RowBox[{"f2", "[", 
            RowBox[{
             RowBox[{"a1", "+", 
              RowBox[{"j1", "*", "dx1"}]}], ",", "a2"}], "]"}], "/", "2"}]}], 
         ";", "\[IndentingNewLine]", 
         RowBox[{"For", "[", 
          RowBox[{
           RowBox[{"j2", "=", "1"}], ",", 
           RowBox[{"j2", "<", 
            RowBox[{"2", "^", "n2"}]}], ",", 
           RowBox[{"j2", "+=", "1"}], ",", 
           RowBox[{"s", "+=", 
            RowBox[{"f2", "[", 
             RowBox[{
              RowBox[{"a1", "+", 
               RowBox[{"j1", "*", "dx1"}]}], ",", 
              RowBox[{"a2", "+", 
               RowBox[{"j2", "*", "dx2"}]}]}], "]"}]}]}], "]"}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"s", "+=", 
          RowBox[{
           RowBox[{"f2", "[", 
            RowBox[{
             RowBox[{"a1", "+", 
              RowBox[{"j1", "*", "dx1"}]}], ",", "b2"}], "]"}], "/", "2"}]}], 
         ";"}]}], "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
      "\[IndentingNewLine]", 
      RowBox[{"s", "+=", 
       RowBox[{
        RowBox[{"f2", "[", 
         RowBox[{"b1", ",", "a2"}], "]"}], "/", "4"}]}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"For", " ", "[", 
       RowBox[{
        RowBox[{"j2", "=", "1"}], ",", 
        RowBox[{"j2", "<", 
         RowBox[{"2", "^", "n2"}]}], ",", 
        RowBox[{"j2", "+=", "1"}], ",", 
        RowBox[{"s", "+=", 
         RowBox[{
          RowBox[{"f2", "[", 
           RowBox[{"b1", ",", 
            RowBox[{"a2", "+", 
             RowBox[{"j2", "*", "dx2"}]}]}], "]"}], "/", "2"}]}]}], "]"}], 
      ";", "\[IndentingNewLine]", 
      RowBox[{"s", "+=", 
       RowBox[{
        RowBox[{"f2", "[", 
         RowBox[{"b1", ",", "b2"}], "]"}], "/", "4"}]}], ";", 
      "\[IndentingNewLine]", "\[IndentingNewLine]", "\[IndentingNewLine]", 
      RowBox[{"s", "*", "dx1", "*", "dx2"}]}]}], "\[IndentingNewLine]", 
    "]"}]}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"f2", "[", 
     RowBox[{
      RowBox[{"Real", " ", "x_"}], ",", 
      RowBox[{"Real", " ", "y_"}]}], "]"}], "\[Rule]", "Real"}], ":=", 
   RowBox[{
    RowBox[{"x", "^", "2"}], "+", "y"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"BuildCode", "[", "]"}], "\[IndentingNewLine]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "fint", "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`fint}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086620082478533`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "fint", "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`fint}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086620082947283`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "fint", "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`fint}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086620083416023`*^9}],

Cell[BoxData["\<\"Successful compilation to C++: 4 function(s)\"\>"], "Print",
 CellChangeTimes->{3.5086620104978247`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NIntegrate", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"x", "^", "2"}], "+", "y"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "0", ",", "1"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"y", ",", "0", ",", "1"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData["0.8333333333333331`"], "Output",
 CellChangeTimes->{3.508662013982155*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{"InstallCode", "[", "]"}]}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.508662014107154*^9}],

Cell[BoxData[
 RowBox[{"LinkObject", "[", 
  RowBox[{"\<\"\\\"C:\\\\b\\\\omath\\\\MathCode\\\\Demos\\\\NIntegrate\\\\\
Globalml.exe\\\"\"\>", ",", "23", ",", "8"}], "]"}]], "Output",
 CellChangeTimes->{3.508662014107154*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{"funt2a", "[", 
   RowBox[{"0.", ",", "1.", ",", "4", ",", "0.", ",", "1.", ",", "4"}], 
   "]"}]}]], "Input"],

Cell[BoxData["0.833984375`"], "Output",
 CellChangeTimes->{3.5086620141227784`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"funt2a", "[", 
  RowBox[{"0.", ",", "1.", ",", "6", ",", "0.", ",", "1.", ",", "6"}], 
  "]"}]], "Input"],

Cell[BoxData["0.8333740234375`"], "Output",
 CellChangeTimes->{3.5086620141227784`*^9}]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"CleanMathCodeFiles", "[", 
   RowBox[{
    RowBox[{"Confirm", "\[Rule]", "False"}], ",", 
    RowBox[{"CleanAllBut", "\[Rule]", 
     RowBox[{"{", "}"}]}]}], "]"}], ";"}]}], "Input"]
},
ScreenStyleEnvironment->"Working",
WindowToolbars->{},
WindowSize->{678, 586},
WindowMargins->{{63, Automatic}, {Automatic, 128}},
ShowCellTags->True,
FrontEndVersion->"6.0 for Microsoft Windows (32-bit) (April 20, 2007)",
StyleDefinitions->Notebook[{
   Cell[
    CellGroupData[{
      Cell["Style Definitions", "Subtitle"], 
      Cell[
      "\nModify the definitions below to change the default appearance of all \
cells in \na given style.  Make modifications to any definition using \
commands in the \nFormat menu.\n", "Text"], 
      Cell[
       CellGroupData[{
         Cell["Style Environment Names", "Section"], 
         Cell[
          StyleData[All, "Working"], PageWidth -> WindowWidth, ScriptMinSize -> 
          9], 
         Cell[
          StyleData[All, "Presentation"], PageWidth -> WindowWidth, 
          ScriptMinSize -> 12, FontSize -> 16], 
         Cell[
          StyleData[All, "Condensed"], PageWidth -> WindowWidth, 
          CellBracketOptions -> {"Margins" -> {1, 1}, "Widths" -> {0, 5}}, 
          ScriptMinSize -> 8, FontSize -> 11], 
         Cell[
          StyleData[All, "Printout"], PageWidth -> PaperWidth, ScriptMinSize -> 
          7, FontSize -> 10, 
          PrivateFontOptions -> {"FontType" -> "Outline"}]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Notebook Options", "Section"], 
         Cell[
         "\nThe options defined for the style below will be used at the \
Notebook level.\n", "Text"], 
         Cell[
          StyleData["Notebook"], PageHeaders -> {{
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"], None, 
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"]}, {
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"], None, 
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"]}}, 
          PageHeaderLines -> {True, True}, 
          PrintingOptions -> {
           "FacingPages" -> True, "FirstPageHeader" -> False}, 
          CellFrameLabelMargins -> 6, StyleMenuListing -> None]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headings", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 0}, CounterIncrements -> "Title", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subtitle", 0}, {"Subsubtitle", 0}}, FontSize -> 
             34, FontColor -> GrayLevel[1], Background -> RGBColor[0, 0, 1]], 
            
            Cell[
             StyleData["Title", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 0}, FontSize -> 
             44], 
            Cell[
             StyleData["Title", "Condensed"], CellMargins -> {{0, 0}, {0, 0}},
              FontSize -> 20], 
            Cell[
             StyleData["Title", "Printout"], CellMargins -> {{0, 0}, {0, 0}}, 
             FontSize -> 24, FontTracking -> "Plain", Background -> 
             GrayLevel[0]]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 3}, ParagraphIndent -> -96, CounterIncrements -> 
             "Subtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subsubtitle", 0}}, FontFamily -> "Helvetica", 
             FontSize -> 18, FontColor -> GrayLevel[1], Background -> 
             RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Subtitle", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 6}, 
             ParagraphIndent -> -157, FontSize -> 30], 
            Cell[
             StyleData["Subtitle", "Condensed"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -78, 
             FontSize -> 14], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -85, 
             FontSize -> 16, Background -> GrayLevel[0.6]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {30, 10}}, PageBreakBelow -> False, 
             CounterIncrements -> "Subsubtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}}, FontFamily -> "Helvetica", FontSize -> 14, 
             FontSlant -> "Italic"], 
            Cell[
             StyleData["Subsubtitle", "Presentation"], 
             CellMargins -> {{8, 10}, {40, 20}}, LineSpacing -> {1, 0}, 
             FontSize -> 24], 
            Cell[
             StyleData["Subsubtitle", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 8}}, FontSize -> 12], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{9, 10}, {50, 10}}, FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], CellFrame -> {{6, 0}, {0, 1}}, CellDingbat -> 
             None, CellMargins -> {{12, Inherited}, {4, 24}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameMargins -> 6, CounterIncrements -> "Section", 
             CounterAssignments -> {{"Subsection", 0}, {"Subsubsection", 0}}, 
             FontFamily -> "Helvetica", FontSize -> 16, FontWeight -> "Bold", 
             FontColor -> RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Section", "Presentation"], 
             CellMargins -> {{10, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["Section", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Section", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, FontSize -> 14, FontTracking -> 
             "Plain", FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], 
             CellMargins -> {{12, Inherited}, {8, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 40}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsection", "Presentation"], 
             CellMargins -> {{11, 10}, {8, 32}}, LineSpacing -> {1, 0}, 
             FontSize -> 22], 
            Cell[
             StyleData["Subsection", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Subsection", "Printout"], 
             CellMargins -> {{9, 0}, {4, 40}}, FontSize -> 12]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> "\[FilledSquare]", 
             CellMargins -> {{25, Inherited}, {8, 12}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsubsection", "Presentation"], 
             CellMargins -> {{29, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Subsubsection", "Condensed"], 
             CellMargins -> {{22, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Subsubsection", "Printout"], 
             CellMargins -> {{21, 0}, {4, 20}}, FontSize -> 11]}, Open]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Body Text", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Text"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Text", FontFamily -> "Times"], 
            Cell[
             StyleData["Text", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Text", "Condensed"], CellMargins -> {{8, 10}, {4, 4}},
              LineSpacing -> {1, 1}, ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Text", "Printout"], CellMargins -> {{9, 0}, {4, 4}}, 
             ParagraphSpacing -> {0, 6}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SmallText"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 6}, 
             CounterIncrements -> "SmallText", FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["SmallText", "Condensed"], 
             CellMargins -> {{8, 10}, {2, 2}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Printout"], 
             CellMargins -> {{9, 0}, {4, 4}}, FontSize -> 7]}, Closed]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Input/Output", "Section"], 
         Cell[
         "\nThe cells in this section define styles used for input and output \
to the \nkernel.  Be careful when modifying, renaming, or removing these \
styles, \nbecause the front end associates special meanings with these style \
names.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Input"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             CounterIncrements -> "Input", FontWeight -> "Bold", Background -> 
             RGBColor[1, 1, 0]], 
            Cell[
             StyleData["Input", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Input", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["Input", "Printout"], CellMargins -> {{44, 0}, {6, 6}},
              Background -> GrayLevel[0.8]]}, Closed]], 
         Cell[
          StyleData["InlineInput"], Evaluatable -> True, CellGroupingRules -> 
          "InputGrouping", CellHorizontalScrolling -> True, PageBreakWithin -> 
          False, GroupPageBreakWithin -> False, DefaultFormatType -> 
          DefaultInputFormatType, AutoItalicWords -> {}, FormatType -> 
          InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
          CounterIncrements -> "Input", FontWeight -> "Bold"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Output"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, CellEditDuplicate -> True, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, 
             CellLabelMargins -> {{3, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Output", Background -> 
             RGBColor[1, 0.900008, 0.900008]], 
            Cell[
             StyleData["Output", "Presentation"], 
             CellMargins -> {{62, Inherited}, {12, 5}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Output", "Condensed"], 
             CellMargins -> {{40, Inherited}, {4, 1}}], 
            Cell[
             StyleData["Output", "Printout"], 
             CellMargins -> {{44, 0}, {6, 2}}, Background -> 
             GrayLevel[0.900008]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Message"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Message", StyleMenuListing -> 
             None, FontColor -> RGBColor[1, 0, 0]], 
            Cell[
             StyleData["Message", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Message", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Message", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Print"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, ShowCellLabel -> 
             False, DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Print", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Print", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Print", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Print", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Graphics"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "GraphicsGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Graphics", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Graphics", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Graphics", "Condensed"], 
             CellMargins -> {{52, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {175, 175}], 
            Cell[
             StyleData["Graphics", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {250, 250}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["CellLabel"], StyleMenuListing -> None, FontFamily -> 
             "Helvetica", FontSize -> 11, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["CellLabel", "Presentation"], FontSize -> 12], 
            Cell[
             StyleData["CellLabel", "Condensed"], FontSize -> 8], 
            Cell[
             StyleData["CellLabel", "Printout"], FontSize -> 8, FontColor -> 
             GrayLevel[0]]}, Closed]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Unique Styles", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {2, 10}}, FontSize -> 16, FontSlant -> 
             "Italic"], 
            Cell[
             StyleData["Author", "Presentation"], 
             CellMargins -> {{12, 10}, {2, 12}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}, FontSize -> 20], 
            Cell[
             StyleData["Author", "Condensed"], 
             CellMargins -> {{8, 10}, {1, 4}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}, FontSize -> 12], 
            Cell[
             StyleData["Author", "Printout"], 
             CellMargins -> {{9, 0}, {4, 12}}, ParagraphSpacing -> {0, 6}, 
             FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Copyright"], ShowCellBracket -> False, 
             CellMargins -> {{10, 10}, {40, 2}}, FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Presentation"], 
             CellMargins -> {{12, 10}, {50, 2}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["Copyright", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 1}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Printout"], 
             CellMargins -> {{9, 0}, {72, 4}}, FontSize -> 7]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Example"], CellMargins -> {{12, 10}, {5, 12}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Example", FontFamily -> "Times", 
             FontWeight -> "Bold"], 
            Cell[
             StyleData["Example", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 20}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Example", "Condensed"], 
             CellMargins -> {{8, 10}, {4, 8}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Example", "Printout"], 
             CellMargins -> {{9, 0}, {4, 10}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Definition"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Definition", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Definition", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Definition", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Definition", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.6]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Theorem"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Theorem", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Theorem", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Theorem", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Theorem", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.4]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseMain"], CellFrame -> {{6, 0}, {0, 1}}, 
             CellMargins -> {{12, Inherited}, {4, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameColor -> RGBColor[0.571389, 0.19675, 0.570504], 
             CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 16, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["ExerciseMain", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["ExerciseMain", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["ExerciseMain", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, CellFrameColor -> 
             GrayLevel[0.500008], FontSize -> 14, FontTracking -> "Plain", 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Exercise"], CellDingbat -> "\[FilledDownTriangle]", 
             CellMargins -> {{23, Inherited}, {4, 18}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Exercise", "Presentation"], 
             CellMargins -> {{33, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Exercise", "Condensed"], 
             CellMargins -> {{17, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Exercise", "Printout"], 
             CellFrame -> {{0, 0}, {0.5, 0}}, CellDingbat -> None, 
             CellMargins -> {{9, 0}, {6, 20}}, FontSize -> 11, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseText"], CellMargins -> {{62, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 8}, 
             CounterIncrements -> "ExcersiceText", FontFamily -> "Times"], 
            Cell[
             StyleData["ExerciseText", "Presentation"], 
             CellMargins -> {{74, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["ExerciseText", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 2}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["ExerciseText", "Printout"], 
             CellMargins -> {{54, 0}, {4, 4}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Itemize"], CellDingbat -> "\[FilledSmallCircle]", 
             CellMargins -> {{65, Inherited}, {8, 8}}, FontFamily -> "Times", 
             FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Itemize", "Presentation"], 
             CellMargins -> {{Inherited, 10}, {8, 12}}, LineSpacing -> {1, 0},
              FontSize -> 18], 
            Cell[
             StyleData["Itemize", "Condensed"], 
             CellMargins -> {{Inherited, Inherited}, {2, 12}}, FontSize -> 
             10], 
            Cell[
             StyleData["Itemize", "Printout"], 
             CellMargins -> {{Inherited, 0}, {4, 20}}, FontSize -> 11]}, 
           Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayCode"], CellFrame -> 2, 
             CellMargins -> {{52, 10}, {8, 8}}, CellHorizontalScrolling -> 
             True, PageBreakWithin -> True, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             FontWeight -> "Bold", Background -> RGBColor[1, 1, 0]], 
            Cell[
             StyleData["DisplayCode", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["DisplayCode", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["DisplayCode", "Printout"], 
             CellMargins -> {{44, 0}, {6, 6}}, Background -> GrayLevel[0.8]]},
            Closed]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Formulas and Programming", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayFormula"], CellMargins -> {{62, 10}, {2, 10}}, 
             CellHorizontalScrolling -> True, ScriptLevel -> 0, 
             SingleLetterItalics -> True, 
             UnderoverscriptBoxOptions -> {LimitsPositioning -> True}], 
            Cell[
             StyleData["DisplayFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 18], 
            Cell[
             StyleData["DisplayFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["DisplayFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ChemicalFormula"], CellMargins -> {{62, 10}, {2, 10}},
              AutoSpacing -> False, ScriptBaselineShifts -> {0.6, Automatic}, 
             ScriptLevel -> 1, SingleLetterItalics -> False, ZeroWidthTimes -> 
             True], 
            Cell[
             StyleData["ChemicalFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Program"], CellMargins -> {{12, 10}, {Inherited, 6}}, 
             FontFamily -> "Courier"], 
            Cell[
             StyleData["Program", "Presentation"], 
             CellMargins -> {{13, 30}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Condensed"], 
             CellMargins -> {{8, 10}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Printout"], 
             CellMargins -> {{9, 0}, {Inherited, 4}}, FontSize -> 9.5]}, 
           Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Automatic Numbering", "Section"], 
         Cell[
         "\nThe following styles are useful for numbered equations, figures, \
etc.  They \nautomatically give the cell a FrameLabel containing a reference \
to a \nparticular counter, and also increment that counter.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedEquation"], 
             CellMargins -> {{62, 10}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, 
                Cell[
                 TextData[{"(", 
                   CounterBox["NumberedEquation"], ")"}]]}, {None, None}}, 
             DefaultFormatType -> DefaultInputFormatType, CounterIncrements -> 
             "NumberedEquation", FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedEquation", "Presentation"], 
             CellMargins -> {{74, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Condensed"], 
             CellMargins -> {{52, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Printout"], 
             CellMargins -> {{54, 0}, {Inherited, Inherited}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedFigure"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Figure ", 
                   CounterBox["NumberedFigure"]}]], None}}, CounterIncrements -> 
             "NumberedFigure", 
             ImageMargins -> {{43, Inherited}, {Inherited, 0}}, 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedFigure", "Presentation"]], 
            Cell[
             StyleData["NumberedFigure", "Condensed"]], 
            Cell[
             StyleData["NumberedFigure", "Printout"]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedTable"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Table ", 
                   CounterBox["NumberedTable"]}]], None}}, TextAlignment -> 
             Center, CounterIncrements -> "NumberedTable", 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedTable", "Presentation"]], 
            Cell[
             StyleData["NumberedTable", "Condensed"]], 
            Cell[
             StyleData["NumberedTable", "Printout"]]}, Open]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headers and Footers", "Section"], 
         Cell[
          StyleData["Header"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontSlant -> "Italic"], 
         Cell[
          StyleData["Footer"], CellMargins -> {{0, 0}, {0, 4}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 6],
          
         Cell[
          StyleData["PageNumber"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontWeight -> "Bold"]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Hyperlink Styles", "Section"], 
         Cell[
         "\nThe cells below define styles useful for making hypertext \
ButtonBoxes.  The \n\"Hyperlink\" style is for links within the same \
Notebook, or between \nNotebooks.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Hyperlink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, 
             FontVariations -> {"Underline" -> True}, FontColor -> 
             RGBColor[0, 0, 1], Background -> None, 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`NotebookLocate[#2]}]& ), ButtonNote -> 
               ButtonData}], 
            Cell[
             StyleData["Hyperlink", "Presentation"]], 
            Cell[
             StyleData["Hyperlink", "Condensed"]], 
            Cell[
             StyleData["Hyperlink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Closed]], 
         Cell[
         "\nThe following styles are for linking automatically to the on-line \
help \nsystem.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["MainBookLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["MainBook", #]}]& )}], 
            Cell[
             StyleData["MainBookLink", "Presentation"]], 
            Cell[
             StyleData["MainBookLink", "Condensed"]], 
            Cell[
             StyleData["MainBookLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["AddOnsLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["AddOns", #]}]& )}], 
            Cell[
             StyleData["AddOnsLink", "Presentation"]], 
            Cell[
             StyleData["AddOnsLink", "Condensed"]], 
            Cell[
             StyleData["AddOnLink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["RefGuideLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["RefGuideLink", #]}]& )}], 
            Cell[
             StyleData["RefGuideLink", "Presentation"]], 
            Cell[
             StyleData["RefGuideLink", "Condensed"]], 
            Cell[
             StyleData["RefGuideLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["GettingStartedLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["GettingStarted", #]}]& )}], 
            Cell[
             StyleData["GettingStartedLink", "Presentation"]], 
            Cell[
             StyleData["GettingStartedLink", "Condensed"]], 
            Cell[
             StyleData["GettingStartedLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["OtherInformationLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["OtherInformation", #]}]& )}], 
            Cell[
             StyleData["OtherInformationLink", "Presentation"]], 
            Cell[
             StyleData["OtherInformationLink", "Condensed"]], 
            Cell[
             StyleData["OtherInformationLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]]}, Open]]}, 
     Open]]}, Visible -> False, FrontEndVersion -> 
  "6.0 for Microsoft Windows (32-bit) (April 20, 2007)", StyleDefinitions -> 
  "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[580, 21, 43, 0, 55, "Title"],
Cell[626, 23, 142, 5, 85, "Text"],
Cell[771, 30, 38, 0, 37, "Subtitle"],
Cell[812, 32, 58, 0, 25, "Text"],
Cell[873, 34, 192, 6, 48, "Input"],
Cell[1068, 42, 67, 0, 25, "Text"],
Cell[1138, 44, 259, 8, 48, "Input"],
Cell[1400, 54, 40, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[1465, 58, 81, 2, 48, "Input"],
Cell[1549, 62, 106, 2, 47, "Output"]
}, Open  ]],
Cell[1670, 67, 204, 4, 43, "Text"],
Cell[1877, 73, 1252, 35, 209, "Input"],
Cell[3132, 110, 159, 4, 73, "Text"],
Cell[CellGroupData[{
Cell[3316, 118, 102, 2, 48, "Input"],
Cell[3421, 122, 104, 2, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3562, 129, 371, 10, 48, "Input"],
Cell[3936, 141, 90, 1, 47, "Output"]
}, Open  ]],
Cell[4041, 145, 46, 0, 37, "Subtitle"],
Cell[CellGroupData[{
Cell[4112, 149, 92, 2, 48, "Input"],
Cell[4207, 153, 450, 10, 23, "Print"]
}, Open  ]],
Cell[4672, 166, 116, 3, 48, "Input"],
Cell[4791, 171, 296, 9, 69, "Input"],
Cell[5090, 182, 1494, 42, 229, "Input"],
Cell[CellGroupData[{
Cell[6609, 228, 75, 2, 48, "Input"],
Cell[6687, 232, 316, 8, 23, "Print"],
Cell[7006, 242, 314, 8, 23, "Print"],
Cell[7323, 252, 316, 8, 23, "Print"],
Cell[7642, 262, 120, 1, 23, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[7799, 268, 114, 3, 69, "Input"],
Cell[7916, 273, 226, 5, 23, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[8179, 283, 142, 4, 69, "Input"],
Cell[8324, 289, 106, 2, 47, "Output"]
}, Open  ]],
Cell[8445, 294, 298, 8, 69, "Input"],
Cell[8746, 304, 52, 0, 37, "Subtitle"],
Cell[8801, 306, 486, 16, 121, "Text"],
Cell[CellGroupData[{
Cell[9312, 326, 294, 7, 48, "Input"],
Cell[9609, 335, 106, 2, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[9752, 342, 345, 8, 69, "Input"],
Cell[10100, 352, 764, 13, 56, "Message"],
Cell[10867, 367, 103, 2, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[11007, 374, 345, 8, 69, "Input"],
Cell[11355, 384, 764, 13, 56, "Message"],
Cell[12122, 399, 104, 2, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[12263, 406, 388, 9, 89, "Input"],
Cell[12654, 417, 766, 13, 56, "Message"],
Cell[13423, 432, 106, 2, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[13566, 439, 388, 9, 89, "Input"],
Cell[13957, 450, 764, 13, 56, "Message"],
Cell[14724, 465, 104, 2, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[14865, 472, 388, 9, 89, "Input"],
Cell[15256, 483, 764, 13, 56, "Message"],
Cell[16023, 498, 104, 2, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[16164, 505, 388, 9, 89, "Input"],
Cell[16555, 516, 104, 2, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[16696, 523, 388, 9, 89, "Input"],
Cell[17087, 534, 104, 2, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[17228, 541, 388, 9, 89, "Input"],
Cell[17619, 552, 106, 2, 47, "Output"]
}, Open  ]],
Cell[17740, 557, 427, 15, 73, "Text"],
Cell[18170, 574, 71, 0, 37, "Subtitle"],
Cell[18244, 576, 4271, 119, 509, "Input"],
Cell[22518, 697, 269, 9, 48, "Input"],
Cell[CellGroupData[{
Cell[22812, 710, 93, 2, 69, "Input"],
Cell[22908, 714, 316, 8, 23, "Print"],
Cell[23227, 724, 316, 8, 23, "Print"],
Cell[23546, 734, 316, 8, 23, "Print"],
Cell[23865, 744, 122, 1, 23, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[24024, 750, 263, 8, 48, "Input"],
Cell[24290, 760, 88, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[24415, 766, 95, 2, 69, "Input"],
Cell[24513, 770, 226, 5, 23, "Print"],
Cell[24742, 777, 224, 4, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[25003, 786, 167, 4, 69, "Input"],
Cell[25173, 792, 83, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[25293, 798, 129, 3, 48, "Input"],
Cell[25425, 803, 87, 1, 47, "Output"]
}, Open  ]],
Cell[25527, 807, 298, 8, 69, "Input"]
}
]
*)

(* End of internal cache information *)
